// I don't fully understand the image moments calculator, so i wrote this test
// program.
//
// My question is: What exactly are the principal moments that it returns? I
// guessed it was the variance around the mean when all voxels are projected
// onto the principal axis, but the values were much larger than I expected.


#include <iostream>
#include <vector>
#include <cmath>
#include "itkImage.h"
#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkResampleImageFilter.h"
#include "itkNearestNeighborInterpolateImageFunction.h"
#include "itkImageMomentsCalculator.h"

using std::cout;
using std::endl;

int main(int argc, char *argv[]) {
  if (argc != 2) {
    std::cerr << "Usage: " << argv[0] << " <2d image file>\n"
      << "\n"
      << "The program computes image moments before and after PCA.\n"
      << "After PCA, the variance of the foreground pixel indices is computed\n"
      << "as well.\n";
    exit(1);
  }
  
  const int Dimension = 2;
  typedef itk::Image<unsigned short, Dimension> ImageType;

  // Read the image. In my application, this will be binary.
  typedef itk::ImageFileReader<ImageType> ImageReaderType;
  ImageReaderType::Pointer reader = ImageReaderType::New();
  reader->SetFileName(argv[1]);
  reader->Update();

  // Now calculate its moments
  typedef itk::ImageMomentsCalculator<ImageType> MomentsCalulatorType;
  MomentsCalulatorType::Pointer moments = MomentsCalulatorType::New();
  moments->SetImage(reader->GetOutput());
  moments->Compute();

  cout << "\n\nMoments before principal axis transformation" << endl;
  moments->Print(cout);

  // Set up a transformation that aligns the principal axes with the
  // coordinate system axes by a rotation around the center of mass
  typedef itk::NearestNeighborInterpolateImageFunction<ImageType, double> InterpolatorType;
  InterpolatorType::Pointer interpolator = InterpolatorType::New();

  typedef MomentsCalulatorType::AffineTransformType TransformType;
  TransformType::Pointer momentstransform = moments->GetPrincipalAxesToPhysicalAxesTransform();
  TransformType::Pointer transform = TransformType::New();
  transform->SetMatrix(momentstransform->GetMatrix());
  TransformType::InputPointType center;
  for (int i = 0; i < Dimension; ++i)
    center[i] = momentstransform->GetOffset()[i];
  transform->SetCenter(center);
  TransformType::OutputVectorType translation;
  translation.Fill(0.0);
  transform->SetTranslation(translation);

  // Now all is set for resampling
  typedef itk::ResampleImageFilter<ImageType, ImageType> ResampleFilterType;
  ResampleFilterType::Pointer resample = ResampleFilterType::New();
  resample->SetInput(reader->GetOutput());
  resample->SetOutputParametersFromImage(reader->GetOutput());
  resample->SetInterpolator(interpolator);
  resample->SetDefaultPixelValue(0);
  resample->SetTransform(transform);
  resample->Update();

  // Write resampled image
  typedef itk::ImageFileWriter<ImageType> WriterType;
  WriterType::Pointer writer = WriterType::New();
  writer->SetInput(resample->GetOutput());
  writer->SetFileName("resampled.hdr");
  writer->Update();

  // Calculate the moments of the resampled image
  MomentsCalulatorType::Pointer moments2 = MomentsCalulatorType::New();
  moments2->SetImage(resample->GetOutput());
  moments2->Compute();

  cout << "\n\nMoments after principal axis transformation" << endl;
  moments2->Print(cout);

  // Calculate the variance of the coordinates
  cout << "\n\nManual calculation of moments after principal axis transformation" << endl;
  std::vector<double> mean;
  mean.push_back(0.0);
  mean.push_back(0.0);

  double numberOfVoxels = 0;

  typedef itk::ImageRegionConstIterator<ImageType> IterType;
  IterType iter(resample->GetOutput(), resample->GetOutput()->GetLargestPossibleRegion());

  iter.GoToBegin();
  while (!iter.IsAtEnd()) {
    if (iter.Get() != 0) {
      mean[0] += iter.GetIndex()[0];
      mean[1] += iter.GetIndex()[1];
      numberOfVoxels++;
    }
    ++iter;
  }
  mean[0] /= numberOfVoxels;
  mean[1] /= numberOfVoxels;
  cout << "Number of Voxels: " << numberOfVoxels << endl;
  cout << "Mean Index: " << mean[0] << ", " << mean[1] <<  endl;

  std::vector<double> variance;
  variance.push_back(0.0);
  variance.push_back(0.0);

  iter.GoToBegin();
  while (!iter.IsAtEnd()) {
    if (iter.Get() != 0) {
      variance[0] += (mean[0]-iter.GetIndex()[0])*(mean[0]-iter.GetIndex()[0]);
      variance[1] += (mean[0]-iter.GetIndex()[1])*(mean[0]-iter.GetIndex()[1]);
    }
    ++iter;
  }
  variance[0] /= numberOfVoxels;
  variance[1] /= numberOfVoxels;
  cout << "Index variance: " << variance[0] << ", " << variance[1]  << endl;

  return 0;
}
